from typing import Any
from saboteur.domain.mutation.strategies import MutationStrategy
from saboteur.domain.mutation.contexts import MutationContext


class TypeFlipStrategy(MutationStrategy):
    def is_applicable(self, context: MutationContext) -> bool:
        return isinstance(context.original_value, (int, str))

    def apply(self, context: MutationContext) -> Any:
        val = context.original_value
        if isinstance(val, int):
            return str(val)
        if isinstance(val, str):
            return int(val) if val.isdigit() else -1


import copy
import random

from typing import List, Dict, Any
from saboteur.domain.mutation.strategies import MutationStrategy
from saboteur.domain.mutation.contexts import MutationContext
from saboteur.domain.mutation.configs import MutationConfig


class Saboteur:
    def __init__(self, config: MutationConfig):
        self.__config = config
    
    def _get_applicable_strategies(self, context: MutationContext) -> List[MutationStrategy]:
        return [s for s in self.__config.strategies if s.is_applicable(context)]
    
    def _mutate(self, context: MutationContext) -> List[object]:
        mutated = []
        for strategy in self._get_applicable_strategies(context):
            mutated_value = strategy.apply(context)
            mutated.append(mutated_value)
        return mutated
    
    def _wrap_into_mutations(self, data: Dict[str, Any]) -> List[MutationContext]:
        if self.__config.apply_all_keys:
            return [
                MutationContext(
                    path=key,
                    original_value=value,
                    original_type=type(value)
                )
                for key, value in data.items()
            ]
        else:
            key = random.choice(list(data.keys()))
            value = data[key]
            return [
                MutationContext(
                    path=key,
                    original_value=value,
                    original_type=type(value)
                )
            ] 
    
    def attack(self, data: Dict[str, Any]) -> Dict[str, Any]:
        _data = copy.deepcopy(data)
        contexts = self._wrap_into_mutations(_data)
        
        for context in contexts:
            candidates = self._get_applicable_strategies(context)
            if not candidates:
                continue
            if self.__config.apply_all_strategies:
                for strategy in candidates:
                    mutated_value = strategy.apply(context)
                    _data[context.path] = mutated_value
            else:
                strategy = random.choice(candidates)
                mutated_value = strategy.apply(context)
                _data[context.path] = mutated_value
        
        return _data


from typing import Any
from saboteur.domain.mutation.strategies import MutationStrategy
from saboteur.domain.mutation.contexts import MutationContext


class NullInjectionStrategy(MutationStrategy):
    def is_applicable(self, context: MutationContext) -> bool:
        return context.original_value is not None

    def apply(self, context: MutationContext) -> Any:
        return None


from typing import Optional, List
from pydantic import (
    BaseModel,
    Field,
    model_validator,
)

from saboteur.domain.mutation.strategies import MutationStrategy


class MutationConfig(BaseModel):
    model_config = {"arbitrary_types_allowed": True}
    
    strategies: List[MutationStrategy] = Field(default_factory=list, description="List of mutation strategies to be used.")
    apply_all_strategies: bool = Field(default=False, description="Flag to determine if all applicable strategies should be applied.")
    num_strategies: Optional[int] = Field(default=None, description="Number of strategies to apply randomly if not applying all.")
    apply_all_keys: bool = Field(default=False, description="Flag to determine if mutations should be applied to all keys in the data structure.")
    num_keys: Optional[int] = Field(default=None, description="Number of keys to mutate randomly if not applying to all keys.")
    
    @model_validator(mode="after")
    def validate_strategy_counts(self):
        if self.apply_all_strategies:
            assert self.num_strategies is None, (
                "num_strategies should be None when apply_all_strategies is True."
            )
        else:
            assert self.num_strategies is not None or self.num_strategies > 0, (
                "num_strategies should be a positive integer or not None when apply_all_strategies is False."
            )
        return self
    
    @model_validator(mode="after")
    def validate_key_counts(self):
        if self.apply_all_keys:
            assert self.num_keys is None, (
                "num_keys should be None when apply_all_keys is True."
            )
        else:
            assert self.num_keys is not None or self.num_keys > 0, (
                "num_keys should be a positive integer or not None when apply_all_keys is False."
            )
        return self


from dataclasses import dataclass
from typing import TypeVar, Generic, Type


T = TypeVar('T')

@dataclass(frozen=True)
class MutationContext(Generic[T]):
    path: str
    original_value: T
    original_type: Type[T]


from abc import ABC, abstractmethod
from saboteur.domain.mutation.contexts import MutationContext


class MutationStrategy(ABC):
    @abstractmethod
    def is_applicable(self, context: MutationContext) -> bool:
        pass
    
    @abstractmethod
    def apply(self, context: MutationContext) -> object:
        pass